----------------------------------------------------------------------------------
-- COPYRIGHT 2019 Jess Eduardo Mndez Rosales
--This program is free software: you can redistribute it and/or modify
--it under the terms of the GNU General Public License as published by
--the Free Software Foundation, either version 3 of the License, or
--(at your option) any later version.
--
--This program is distributed in the hope that it will be useful,
--but WITHOUT ANY WARRANTY; without even the implied warranty of
--MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
--GNU General Public License for more details.
--
--You should have received a copy of the GNU General Public License
--along with this program.  If not, see <http://www.gnu.org/licenses/>.
--
--
-- Description: Las caracteristicas de este modulo son:
--					* Direccin esclava de 7 bits configurable.
--             * El bit INI debe ponerse a '1' para iniciar lectura o escritura.
--             * El proceso de lectura pone el apuntador en la direccin de registro especificada mediante la direccin esclava+Escritura,
--					  para despus reiniciar la interfaz y mandar la direccin esclava+Lectura.
--             * La frecuencia de reloj del I2C a 1Mhz.
--
--          PROCESO DE ESCRITURA.
--
--
-- INI ____/\__
--
--                           Dir. Esclava + Escritura
--              |--------------------0xD0-----------------------|     
-- SDA \__/\_____/\_______________________|-ACK-|--Dir.  Registro--|-ACK-|-------Dato-------|-ACK-|_____/
--
--	SCL \__/\__/\__/\__/\__/\__/\__/\__/\__/\__/\__ ... __/\__/\__/\__ ... __/\__/\__/
--
-- FIN _________________________________________________________________________________________________________________________/\__
--
--
--
--          PROCESO DE LECTURA.
--
--
-- INI ____/\__
--
--                           Dir. Esclava + Escritura                                                                    Dir. Esclava + Lectura
--              |--------------------0xD0-----------------------|                              |--Reinicio--|--------------------0xD1-----------------------
--	SDA \__/\_____/\_______________________|-ACK-|--Dir.  Registro--|-ACK-|____/\____/\_____/\_________________/|-ACK-|--Dato  in--|-NACK-|____/
--
--	SCL \__/\__/\__/\__/\__/\__/\__/\__/\__/\__/\__ ... __/\__/\__/\____/\__/\__/\__/\__/\__/\__/\__/\__/\__ ... __/\__/\__/ 
--
-- FIN _____________________________________________________________________________________________________________________________________________________________________________________/\__

----------------------------------------------------------------------------------


library IEEE;
use IEEE.STD_LOGIC_1164.ALL;

entity I2C_DS1307_RevA is

GENERIC(
        CLK_FPGA : INTEGER := 50_000_000; --Frecuencia de reloj base.
        DIR_ESC  : STD_LOGIC_VECTOR(7 DOWNTO 0) := X"68" --Direccin esclava de DS1307.
);

PORT(
     CLK      : IN    STD_LOGIC; --Reloj principal.
     SDA_DS   : INOUT STD_LOGIC; --Puerto bidireccional de datos.
     SCL_DS   : OUT   STD_LOGIC; --Puerto de seal de reloj.
     INI      : IN    STD_LOGIC; --Bit que inicia transmisin/recepcin.
     FIN      : OUT   STD_LOGIC; --Bit que indica cundo finaliz la transmisin/recepcin.
     REG      : IN    STD_LOGIC_VECTOR(7 DOWNTO 0); --Direccin del registro a leer o escribir del DS1307.
     DATA_IN  : IN    STD_LOGIC_VECTOR(7 DOWNTO 0); --Dato de entrada a escribir en el registro.
     DATA_OUT : OUT   STD_LOGIC_VECTOR(7 DOWNTO 0); --Byte recibido en la lectura.
     RW       : IN    STD_LOGIC --Bit que configura lectura ('1') o escritura ('0').

);


end I2C_DS1307_RevA;

architecture Behavioral of I2C_DS1307_RevA is

CONSTANT DIR_ESC_R : STD_LOGIC_VECTOR(7 DOWNTO 0) := DIR_ESC(6 DOWNTO 0)&'1'; --Direccin esclava en modo lectura.
CONSTANT DIR_ESC_W : STD_LOGIC_VECTOR(7 DOWNTO 0) := DIR_ESC(6 DOWNTO 0)&'0'; --Direccin esclava en modo escritura.
CONSTANT CONTA_SCL_FIN : INTEGER := ((CLK_FPGA / 1_000_000)/2) - 1; --Valor mximo para generar seal de SCL.
CONSTANT CONTA_GO_FIN  : INTEGER := ((CLK_FPGA / 1_000_000)/2) - 1; --Valor mximo para generar condicin de inicio para SDA
CONSTANT ESC_COND_FIN  : INTEGER := ((CLK_FPGA / 1_000_000)/2) - 1; --Valor mximo para generar condicin de paro para SDA.
CONSTANT ESC_RESTART   : INTEGER := ((CLK_FPGA / 1_000_000)/2) - 1; --Valor mximo para generar condicin de reinicio.

signal en_ai          : std_logic := '0'; --Bit que indica cundo habilitar el ACK.
signal scl_s          : std_logic := '1'; --Seal auxiliar para SCL.
signal sda_s          : std_logic := 'Z'; --Seal auxiliar para SDA (valores lgicos de '0' y '1').
signal sda_s2         : std_logic := 'Z'; --Seal auxiliar para SDA (Valores lgicos de '0' y 'Z').
signal go_sda         : std_logic := '0'; --Bit que indica cundo debe enviarse la seal de SDA.
signal go_scl         : std_logic := '1'; --Bit que indica cundo debe enviarse la seal de SCL.
signal flanco         : std_logic := '0'; --Bit que indica cundo se detect un flanco negativo.
signal flancop        : std_logic := '0'; --Bit que indica cundo se detect un flanco positivo.
signal data_ins       : std_logic := '0'; --Seal auxiliar para SDA en modo de entrada.
signal go_sda_conta   : std_logic := '0'; --Bit que indica cundo debe iniciar el contador para generar SDA.
signal reg_s          : std_logic_vector(7 downto 0) := (others => '0'); --Seal auxiliar que almacena temporalmente la direccin del registo.
signal data_in_s      : std_logic_vector(7 downto 0) := (others => '0'); --Seal auxiliar que almacena temporalmente el byte recibido.
signal data_out_s     : std_logic_vector(7 downto 0) := (others => '0'); --Seal auxiliar que almacena temporalmente el byte a enviar.
signal reg_flanco     : std_logic_vector(3 downto 0) := (others => '0'); --Registro de corrimiento para detectar flancos negativos.
signal reg_flancop    : std_logic_vector(3 downto 0) := (others => '0'); --Registro de corrimiento para detectar flancos positivos.
signal edo            : integer range 0 to 50 := 0;            --Mquina de estados principal.
signal edoc           : integer range 0 to 50 := 0;            --Mquina de estados para generar seal de reloj SCL.
signal conta_go       : integer range 0 to conta_go_fin := 0;  --Contador que le indica cundo enviar la seal de SDA.
signal conta_scl      : integer range 0 to conta_scl_fin := 0; --Contador que genera la seal SCL.
signal conta_bit      : integer range 7 downto 0 := 7;         --Contador de bits que se envan o reciben.
signal conta_restart  : integer range 0 to esc_restart := 0;   --Contador para generar condicin de reinicio en modo lectura.
signal conta_cond_fin : integer range 0 to esc_cond_fin := 0;  --Contador para generar condicin de paro.

begin

--Proceso que detecta flancos de bajada--
process(clk)
begin
  if rising_edge(clk) then
     reg_flanco <= reg_flanco(2 downto 0)&scl_s;
     if reg_flanco = "1100" then
        flanco <= '1';
     else
        flanco <= '0';
     end if;
  end if;
end process;
        
--Proceso que detecta flancos de subida--		  
process(clk)
begin
  if rising_edge(clk) then
     reg_flancop <= reg_flancop(2 downto 0)&scl_s;
     if reg_flancop = "0011" then
        flancop <= '1';
     else
        flancop <= '0';
     end if;
  end if;
end process;


sda_ds   <= sda_s2 when en_ai= '0' else 'Z'; --Habilita o deshabilita el puerto SDA.
sda_s2   <= '0'   when sda_s = '0' else 'Z'; --Convierte los '1' en 'Z'
SCL_DS   <= scl_s;  --Se enva seal de reloj por el puerto de salida.
data_ins <= SDA_DS; --Se lee el puerto SDA en modo lectura.

--Proceso que genera retardo para condicin de inicio--
process(clk)
begin
    if rising_edge(clk) then
        if go_sda_conta = '1' then
            conta_go <= conta_go+1;
            if conta_go = conta_go_fin then
                conta_go <= 0;
                go_sda <= '1';
            else
                go_sda <= '0';
            end if;
        else
            conta_go <= 0;
            go_sda <= '0';
        end if;
    end if;
end process;

--Proceso que indica cundo se debe enviar la seal de SCL--
process(clk,go_scl)
begin
    if go_scl = '0' then
       edoc <= 0;
       scl_s <= '1';
       conta_scl <= 0;
       
    elsif rising_edge(clk) then
        if edoc = 0 then
            if go_sda = '1' then
                edoc <= 1;
            end if;
            scl_s <= '1';
            
        elsif edoc = 1 then
            conta_scl <= conta_scl+1;
            if conta_scl = conta_scl_fin then
                conta_scl <= 0;
                scl_s <= not scl_s;
            end if;
        
        end if;
    end if;
end process;

--Mquina de estados principal que genera el protocolo I2C--
process(clk)
begin
    if rising_edge(clk) then
--Estado que espera a que el bit de inicio se active y verifica si se desea una lectura o escritura.
--Mientras se permanece en este estado el puerto SDA se pone en alta impedancia.	 
        if edo = 0 then
           if ini = '1' and  rw = '0' then 
              edo <= 1;
              go_sda_conta <= '1';
           elsif ini = '1' and  rw = '1' then
              edo <= 12;
              go_sda_conta <= '1';
           end if;
           fin <= '0';
           en_ai<= '1';
           sda_s <= 'Z';
                  
       ----------Escritura--------------   
--Estados que se encargan de escribir en los registros del sensor.

        elsif edo = 1 then--Genera condicin de inicio y almacena temporalmente el dato a enviar y el registro.
           if go_sda = '1' then
              edo <= 40;
              reg_s <= reg;
              data_in_s <= data_in;
              sda_s <= '0';--condicin de inicio
              en_ai<= '0';
              go_scl <= '1';
           end if;      
           
        
        elsif edo = 40 then--Estado dummy esperando flanco de bajada.
            if flanco = '1' then
                edo <= 2;
            end if;
              
        elsif edo = 2 then--Enva los bits de la direccin esclava.
            sda_s <= dir_esc_w(conta_bit);
            en_ai<= '0';
            if flanco = '1' then
               conta_bit <= conta_bit-1;
               if conta_bit = 0 then
                  conta_bit <= 7;
                  edo <= 3;
              end if;
            end if;
            
        elsif edo = 3 then--ack
            if flanco = '1' then
               edo <= 4;
            end if;
               en_ai<= '1';
               sda_s <= '0';
               
       elsif edo = 4 then--Enva los bits del registro.
            sda_s <= reg_s(conta_bit);
            en_ai<= '0';
            if flanco = '1' then
               conta_bit <= conta_bit-1;
               if conta_bit = 0 then
                  edo <= 7;
                  conta_bit <= 7;
               end if;
           end if;
       
       elsif edo = 7 then--ack 3
            if flanco = '1' then
               edo <= 8;
            end if;
            en_ai<= '1';
               sda_s <= '0';
            
        
       elsif edo = 8 then--Enva los bits del dato.
            sda_s <= data_in_s(conta_bit);
            en_ai<= '0';
            if flanco = '1' then
               conta_bit <= conta_bit-1;
               if conta_bit = 0 then
                  edo <= 9;
                  conta_bit <= 7;
               end if;
           end if;
        
        
        elsif edo = 9 then--ack 4
            if flanco = '1' then
               edo <= 10;
            end if;
            en_ai<= '1';
               sda_s <= '0';            
       
       elsif edo = 10 then--Espera al ltimo flanco de subida.
           if flancop = '1' then
              go_scl <= '0';
              go_sda_conta <= '0';
              edo <= 11;
           end if;
           en_ai<= '0';
              sda_s <= '0';
       
       elsif edo = 11 then--Espera un tiempo para generar condicin de paro.
            conta_cond_fin <= conta_cond_fin+1;
            if conta_cond_fin = esc_cond_fin then
                conta_cond_fin <= 0;
                edo <= 30;
                fin <= '1'; --Indica fin de proceso.
                en_ai<= '1';
                sda_s <= 'Z';
            end if;
       -------------------------------
       
       ----------lectura--------------
       elsif edo = 12 then--Estado para generar condicin de inicio y se carga el registro a leer.
           if go_sda = '1' then
              edo <= 41;
              reg_s <= reg;
              data_in_s <= data_in;
              sda_s <= '0';--condicin de inicio
              en_ai<= '0';
              go_scl <= '1';
           end if;
        
        elsif edo = 41 then--Estado dummy esperando flanco de subida.
            if flanco = '1' then
                edo <= 13;
            end if;
              
        elsif edo = 13 then--Enva los bits de la direccin esclava con bit de escritura.
            sda_s <= dir_esc_w(conta_bit);
            en_ai<= '0';
            if flanco = '1' then
               conta_bit <= conta_bit-1;
               if conta_bit = 0 then
                  conta_bit <= 7;
                  edo <= 14;
              end if;
            end if;
            
        elsif edo = 14 then--ack
            if flanco = '1' then
               edo <= 15;
            end if;
            en_ai<= '1';
               
       elsif edo = 15 then--Enva los bits del registro a leer.
            sda_s <= reg_s(conta_bit);
            en_ai<= '0';
            if flanco = '1' then
               conta_bit <= conta_bit-1;
               if conta_bit = 0 then
                  edo <= 18;
                  conta_bit <= 7;
               end if;
           end if;
       
       elsif edo = 18 then--ack 3
            if flancop = '1' then
               edo <= 19;
            end if;
               en_ai<= '1';
      
       elsif edo = 19 then--Estado dummy esperando flanco de subida.
            if flancop = '1' then
               edo <= 20;
            end if;
               
        elsif edo = 20 then--Contador para generar condicin de reinicio.
			  conta_restart <= conta_restart+1;
            if conta_restart = (esc_restart+1)/2 then
               conta_restart <= 0;
               edo <= 21;
               sda_s <= '0';
               en_ai <= '0';
            end if;
		  
		  elsif edo = 21 then--Estado dummy esperando flanco de bajada.
				if flanco = '1' then
                edo <= 22;
            end if;    
		  
		  elsif edo = 22 then--Enva los bits de la direccin esclava con bit de lectura.		  
            sda_s <= dir_esc_r(conta_bit);
            en_ai<= '0';
            if flanco = '1' then
               conta_bit <= conta_bit-1;
               if conta_bit = 0 then
                  conta_bit <= 7;
                  edo <= 24;
              end if;
            end if;  
         
          elsif edo = 24 then--ack 4
            if flanco = '1' then
               edo <= 25;
            end if;
                en_ai<= '1';
          
          elsif edo = 25 then--Recibe los bits que manda el sensor.
            if flancop = '1' then
               data_out_s(conta_bit) <= data_ins;
               conta_bit <= conta_bit-1;
               if conta_bit = 0 then
                  conta_bit <= 7;
                  edo <= 26;
              end if;
            end if; 
            
          elsif edo = 26 then--Se envia el dato recibido por el puerto "DATA_OUT".
            DATA_OUT <= data_out_s;
            edo <= 27;
          
           elsif edo = 27 then--ack 4
            if flanco = '1' then
               edo <= 28;
            end if;
            en_ai<= '1';            
       
       elsif edo = 28 then--Espera al ltimo flanco de subida.
           if flancop = '1' then
              go_scl <= '0';
              go_sda_conta <= '0';
              edo <= 29;
           end if;
              en_ai<= '0';
              sda_s <= '0';
       
       elsif edo = 29 then--Contador para generar condicin de de paro.
            conta_cond_fin <= conta_cond_fin+1;
            if conta_cond_fin = esc_cond_fin then
                conta_cond_fin <= 0;
                edo <= 30;
                fin <= '1';
                en_ai<= '1';
            end if;
        
        elsif edo = 30 then--dummy 2
            fin <= '0';
            edo <= 0;
               
        end if;
    end if;
end process;

end Behavioral;
