----------------------------------------------------------------------------------
-- COPYRIGHT 2019 Jess Eduardo Mndez Rosales
--This program is free software: you can redistribute it and/or modify
--it under the terms of the GNU General Public License as published by
--the Free Software Foundation, either version 3 of the License, or
--(at your option) any later version.
--
--This program is distributed in the hope that it will be useful,
--but WITHOUT ANY WARRANTY; without even the implied warranty of
--MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
--GNU General Public License for more details.
--
--You should have received a copy of the GNU General Public License
--along with this program.  If not, see <http://www.gnu.org/licenses/>.
--
--
-- Description: Libreria para controlar un reloj de tiempo real DS1307, el cual
--					 utiliza una interfaz I2C para la transferencia de datos.
--              
-- Caractersticas:
--             - Frecuencia de reloj de FPGA configurable.
--             - Frecuencia de I2C a 1Mhz.
--             - Tiempo de carga de datos: 207.3 us.
--             - Tiempo de lectura de datos: 270.7 us.
--             - Los datos de entrada y salida se presentan en formato BCD como 
--               se muestran en la tabla.
--
-- La libreria funciona conforme a la hoja de datos del sensor DS1307, cuya tabla 
-- de registros es la siguiente:
--  ____________________________________________________________________________________
-- | REGISTRO | BIT 7 | BIT 6 | BIT 5 | BIT 4 | BIT 3 | BIT 2 | BIT 1 | BIT 0 |  Rango  |  
-- |------------------------------------------------------------------------------------|
-- |   00h    |  xx   |      10 Segundos      |            Segundos           |  0-59   |
-- |----------|-------|-----------------------|-------------------------------|---------|
-- |   01h    |   0   |      10 Minutos       |            Minutos            |  0-59   |
-- |----------|-------|-----------------------|-------------------------------|---------|
-- |          |       |  12   |10 Hrs |       |                               |  1-12   |
-- |   02h    |   0   |-------|-------|10 Hrs |             Horas             | +AM/PM  |
-- |          |       |  24   | AM/PM |       |                               |  0-23   |
-- |----------|-------|-------|-------|-------|-------------------------------|---------|
-- |   03h    |   0   |   0   |   0   |   0   |   0   |    Da de la semana   |   1-7   |
-- |----------|-------|-------|---------------|-------------------------------|---------|
-- |   04h    |   0   |   0   | 10 da del mes|           Da del mes         |   1-31  |
-- |----------|-------|-------|---------------|-------------------------------|---------|
-- |   05h    |   0   |   0   |   0   |10 Mes |              Meses            |   1-12  |
-- |----------|-------------------------------|-------------------------------|---------|
-- |   06h    |           10 Aos             |               Aos            |   0-99  |
--  
--
-- Por ejemplo, se configurar el sensor con la siguiente informacin:
-- 19/11/2019 (DD/MM/AA) Martes, 1:20:00 PM.
-- 
-- En los puertos de entrada se colocarn los valores de cada parmetro en formato BCD.
--
--   FORMATO_2412  = '1'   - Formato 12 horas.
--	  AM_PM         = '1'   - PM.
--	  DIA_IN        = x"19" - "0001 1001" en BCD.
--	  MES_IN        = x"11" - "0001 0001" en BCD.
--	  YEAR_IN       = x"19" - "0001 1001" en BCD, el '20' lo debe considerar el usuario.
--	  DIA_SEMANA_IN = x"03" - 3 es martes.
--	  HORA_IN       = x"01" - "0000 0001" en BCD.
--	  MINUTO_IN     = x"20" - "0010 0000" en BCD 
--	  SEGUNDO_IN    = x"00" - "0000 0000" en BCD 
-- 
-- IMPORTANTE: Cualquier valor fuera del rango permitido no sera tomado en cuenta poniendo el valor por default.
--
-- Despus de colocar los datos en los puertos de entrada se debe activar el puerto CARGAR_INI para almacenar
-- los datos en el sensor
--
--             CARGAR DATOS.
--
--                                    |-----------------207.3 us------------------|     
--
--	            CLK _/\__/\__/\__/\__/\__/\ ... __/\__/\__/\__/\__
--
--      CARGAR_INI ___________________/ ... \__
--
--      CARGAR_FIN ___________________________________ ... ____________________/\__
--
--    FORMATO_2412 _____________/ ... 
--	          AM_PM _____________/ ... 
--	         DIA_IN <----xxx----><---------------------x"19"-----------------------><-
--	         MES_IN <----xxx----><---------------------x"11"-----------------------><-
--	        YEAR_IN <----xxx----><---------------------x"19"-----------------------><-
--	  DIA_SEMANA_IN <----xxx----><---------------------x"03"-----------------------><-
--	        HORA_IN <----xxx----><---------------------x"01"-----------------------><-
--	      MINUTO_IN <----xxx----><---------------------x"20"-----------------------><-
--	     SEGUNDO_IN <----xxx----><---------------------x"00"-----------------------><-
--
--
--
-- Para leer los datos se debe activar el puerto "LEER_INI" y esperar a que "LEER_FIN" se ponga a '1';
--
--             LEER DATOS.
--
--                                    |-----------------270.7 us------------------|     
--
--	             CLK _/\__/\__/\__/\__/\__/\ ... __/\__/\__/\__/\__/\__/\__/\__/\__
--
--       CARGAR_INI ___________________/ ... \__________________________
--
--       CARGAR_FIN ___________________________________ ... ____________________/\__________________________
--
--	         DIA_OUT <-----------------------------xxx-- ... --------------------><-------DATO VLIDO----------
--	         MES_OUT <-----------------------------xxx-- ... --------------------><-------DATO VLIDO----------
--	        YEAR_OUT <-----------------------------xxx-- ... --------------------><-------DATO VLIDO----------
--	  DIA_SEMANA_OUT <-----------------------------xxx-- ... --------------------><-------DATO VLIDO----------
--	        HORA_OUT <-----------------------------xxx-- ... --------------------><-------DATO VLIDO----------
--	      MINUTO_OUT <-----------------------------xxx-- ... --------------------><-------DATO VLIDO----------
--	     SEGUNDO_OUT <-----------------------------xxx-- ... --------------------><-------DATO VLIDO----------
--
-- IMPORTANTE: El puerto "HORA_OUT" contiene los bits de informacin para el formato de hora y AM o PM.
--
----------------------------------------------------------------------------------
library IEEE;
use IEEE.STD_LOGIC_1164.ALL;

entity RTC_DS1307_REVA is

GENERIC(
        FREC_FPGA : INTEGER := 50_000_000 --Frecuencia del reloj del FPGA en Hertz.
);

PORT(
     CLK            : IN    STD_LOGIC; --Reloj principal.
     SDA_DS         : INOUT STD_LOGIC; --Puerto bidireccional de datos.
	  SCL_DS         : OUT   STD_LOGIC; --Puerto de reloj.
     CARGAR_INI     : IN    STD_LOGIC; --Bit que guarda los datos en el DS1307 cuando se pose a '1'.
	  CARGAR_FIN     : OUT   STD_LOGIC; --Bit que indica cuando ya se cargaron los datos.
	  FORMATO_2412   : IN    STD_LOGIC; --Selector de formato de horas. '0' para 24 Hrs, '1' para 12 Hrs. 
	  AM_PM          : IN    STD_LOGIC; --Selector AM o PM. '0' para AM, '1' para PM. SLO UTILIZAR CUANDO SE SELECCIONA FORMATO 12 HORS.
	  DIA_IN         : IN    STD_LOGIC_VECTOR(7 DOWNTO 0); --Dia del mes, rango de 1-31.
	  MES_IN         : IN    STD_LOGIC_VECTOR(7 DOWNTO 0); --Mes, rango de 1-12.
	  YEAR_IN        : IN    STD_LOGIC_VECTOR(7 DOWNTO 0); --Ao, rango de 0-99.
	  DIA_SEMANA_IN  : IN    STD_LOGIC_VECTOR(7 DOWNTO 0); --Da de la semana, rango 1-7. 1-Domingo, 7-Sbado.
	  HORA_IN        : IN    STD_LOGIC_VECTOR(7 DOWNTO 0); --Hora, rango de 1-12 en formato de 12 Hrs. y rango de 00-23 en formato 24 Hrs.
	  MINUTO_IN      : IN    STD_LOGIC_VECTOR(7 DOWNTO 0); --Minutos, rango de 0-59.
	  SEGUNDO_IN     : IN    STD_LOGIC_VECTOR(7 DOWNTO 0); --Segundos, rango de 0-59.
	  LEER_INI       : IN    STD_LOGIC; --Bit que inicia la lectura de datos.
	  LEER_FIN       : OUT   STD_LOGIC; --Bit que indica cundo finaliz la lectura y la informacin se encuentra en los puertos de salida.
	  DIA_OUT        : OUT   STD_LOGIC_VECTOR(7 DOWNTO 0); --Da ledo en formato BCD.
	  MES_OUT        : OUT   STD_LOGIC_VECTOR(7 DOWNTO 0); --Mes ledo en formato BCD.
	  YEAR_OUT       : OUT   STD_LOGIC_VECTOR(7 DOWNTO 0); --Ao ledo en formato BCD.
	  DIA_SEMANA_OUT : OUT   STD_LOGIC_VECTOR(7 DOWNTO 0); --Da de la semana ledo en formato BCD.
	  HORA_OUT       : OUT   STD_LOGIC_VECTOR(7 DOWNTO 0); --Hora leda en formato BCD.
	  MINUTO_OUT     : OUT   STD_LOGIC_VECTOR(7 DOWNTO 0); --Minuto ledo en formato BCD.
	  SEGUNDO_OUT    : OUT   STD_LOGIC_VECTOR(7 DOWNTO 0)  --Segundo ledo en formato BCD.
);

end RTC_DS1307_REVA;

architecture Behavioral of RTC_DS1307_REVA is

--Declarando al protocolo I2C como componente--
COMPONENT I2C_DS1307_RevA is

GENERIC(
        CLK_FPGA : INTEGER := 50_000_000;
        DIR_ESC  : STD_LOGIC_VECTOR(7 DOWNTO 0) := X"68"
);

PORT(
     CLK      : IN    STD_LOGIC;
     SDA_DS   : INOUT STD_LOGIC;
     SCL_DS   : OUT   STD_LOGIC;
     INI      : IN    STD_LOGIC;
     FIN      : OUT   STD_LOGIC;
     REG      : IN    STD_LOGIC_VECTOR(7 DOWNTO 0);
     DATA_IN  : IN    STD_LOGIC_VECTOR(7 DOWNTO 0);
     DATA_OUT : OUT   STD_LOGIC_VECTOR(7 DOWNTO 0);
     RW       : IN    STD_LOGIC

);

end COMPONENT I2C_DS1307_RevA;

signal ini              : std_logic := '0'; --Seal de inicio de lectura/escritura.
signal fin              : std_logic := '0'; --Bit que indica fin de lectura/escritura.
signal rw               : std_logic := '0'; --Bit que selecciona lectura o escritura-
signal reg              : std_logic_vector(7 downto 0) := (others => '0'); --Direccin de registro a leer o escribir.
signal data_in          : std_logic_vector(7 downto 0) := (others => '0'); --Dato a escribir en el registro seleccionado.
signal data_out         : std_logic_vector(7 downto 0) := (others => '0'); --Dato recibido por I2C.
signal dia_in_s         : std_logic_vector(7 downto 0) := (others => '0'); --Da que se carga en el DS1307.
signal mes_in_s         : std_logic_vector(7 downto 0) := (others => '0'); --Mes que se carga en el DS1307.
signal year_in_s        : std_logic_vector(7 downto 0) := (others => '0'); --Ao que se carga en el DS1307.
signal dia_semana_in_s  : std_logic_vector(7 downto 0) := (others => '0'); --Da de la semana que se carga en el DS1307.
signal hora_in_s        : std_logic_vector(7 downto 0) := (others => '0'); --Hora que se carga en el DS1307.
signal minuto_in_s      : std_logic_vector(7 downto 0) := (others => '0'); --Minuto que se carga en el DS1307.
signal segundo_in_s     : std_logic_vector(7 downto 0) := (others => '0'); --Segundo que se carga en el DS1307.
signal dia_out_s        : std_logic_vector(7 downto 0) := (others => '0'); --Da que se ley del DS1307.
signal mes_out_s        : std_logic_vector(7 downto 0) := (others => '0'); --Mes que se ley del DS1307.
signal year_out_s       : std_logic_vector(7 downto 0) := (others => '0'); --Ao que se ley del DS1307.
signal dia_semana_out_s : std_logic_vector(7 downto 0) := (others => '0'); --Da de la semana que se ley del DS1307.
signal hora_out_s       : std_logic_vector(7 downto 0) := (others => '0'); --Hora que se ley del DS1307.
signal minuto_out_s     : std_logic_vector(7 downto 0) := (others => '0'); --Minuto que se ley del DS1307.
signal segundo_out_s    : std_logic_vector(7 downto 0) := (others => '0'); --Segundo que se ley del DS1307.
signal edo : integer range 0 to 31 := 0; --Seal para la mquina de estados principal.




begin

--Instancia del componente que genera protocolo I2C.
U1: I2C_DS1307_RevA GENERIC MAP( CLK_FPGA => 50_000_000, DIR_ESC => X"68" )
	PORT MAP( CLK, SDA_DS, SCL_DS, ini, fin, reg, data_in, data_out, rw );

--Mandando la informacin leda por los puertos correspondientes.
DIA_OUT        <= dia_out_s;
MES_OUT        <= mes_out_s;
YEAR_OUT       <= year_out_s;
DIA_SEMANA_OUT <= dia_semana_out_s;
HORA_OUT       <= hora_out_s;
MINUTO_OUT     <= minuto_out_s;
SEGUNDO_OUT    <= segundo_out_s;

--Mquina de estados que controla el flujo de datos.
process(CLK)
begin
if rising_edge(CLK) then
--Estado inicial donde se cargan los datos para la escritura o se pasa al estado para realizar lectura.
--Se ajustan los bits para evitar que el DS1307 se 	
	if edo = 0 then
		if CARGAR_INI = '1' then
			segundo_in_s <= '0' & SEGUNDO_IN(6 downto 0);
			minuto_in_s  <= '0' & MINUTO_IN(6 downto 0);
			hora_in_s    <= '0' & ((FORMATO_2412&"000000") OR ('0'&AM_PM&"00000") OR HORA_IN(6 downto 0));
			dia_semana_in_s <= "00000" & DIA_SEMANA_IN(2 downto 0);
			dia_in_s     <= "00" & DIA_IN(5 downto 0);
			mes_in_s     <= "000" & MES_IN(4 downto 0);
         year_in_s     <= YEAR_IN;
			edo <= 1;
		elsif LEER_INI = '1' THEN
			edo <= 17;
		end if;	
--Se cargan el registro y los segundos.
	elsif edo = 1 then
		reg <= x"00";
		data_in <= segundo_in_s;
		rw <= '0';
		edo <= 2;
		
	elsif edo = 2 then
		if fin = '1' then
			ini <= '0';
			edo <= 3;
		else
			ini <= '1';
		end if;
		
--Se cargan el registro y los minutos.
	elsif edo = 3 then
		reg <= x"01";
		data_in <= minuto_in_s;
		rw <= '0';
		edo <= 4;
		
	elsif edo = 4 then
		if fin = '1' then
			ini <= '0';
			edo <= 5;
		else
			ini <= '1';
		end if;	
	
--Se cargan el registro y los minutos.	
	elsif edo = 5 then
		reg <= x"02";
		data_in <= hora_in_s;
		rw <= '0';
		edo <= 6;
		
	elsif edo = 6 then
		if fin = '1' then
			ini <= '0';
			edo <= 7;
		else
			ini <= '1';
		end if;	
		
--Se cargan el registro y el dia de la semana.
	elsif edo = 7 then
		reg <= x"03";
		data_in <= dia_semana_in_s;
		rw <= '0';
		edo <= 8;
		
	elsif edo = 8 then
		if fin = '1' then
			ini <= '0';
			edo <= 9;
		else
			ini <= '1';
		end if;	

--Se cargan el registro y el dia.
	elsif edo = 9 then
		reg <= x"04";
		data_in <= dia_in_s;
		rw <= '0';
		edo <= 10;
		
	elsif edo = 10 then
		if fin = '1' then
			ini <= '0';
			edo <= 11;
		else
			ini <= '1';
		end if;

--Se cargan el registro y el mes.
	elsif edo = 11 then
		reg <= x"05";
		data_in <= mes_in_s;
		rw <= '0';
		edo <= 12;
		
	elsif edo = 12 then
		if fin = '1' then
			ini <= '0';
			edo <= 13;
		else
			ini <= '1';
		end if;

--Se cargan el registro y el ao.
	elsif edo = 13 then
		reg <= x"06";
		data_in <= year_in_s;
		rw <= '0';
		edo <= 14;
		
	elsif edo = 14 then
		if fin = '1' then
			ini <= '0';
			edo <= 15;
			CARGAR_FIN <= '1'; --Bit que activa FIN de configuracin.
		else
			ini <= '1';
		end if;

	elsif edo = 15 then
		CARGAR_FIN <= '0'; --Bit que desactiva FIN de configuracin.
		edo <= 16;
		
	elsif edo = 16 then --Estado dummy para darle al usuario un ciclo de reloj para cargar los siguientes datos.
		edo <= 0;

	-- Leer --
--Se configura el protocolo I2C como lectura.	
	elsif edo = 17 then 
		reg <= x"00";
		rw <= '1';
		edo <= 18;
		
	elsif edo = 18 then
		if fin = '1' then
			segundo_out_s <= data_out; --Se mandan los segundos a la seal de segundos.
			ini <= '0';
			edo <= 19;
		else
			ini <= '1';
		end if;
		
--Se configura el protocolo I2C como lectura y se adquieren los minutos.	
	elsif edo = 19 then
		reg <= x"01";
		rw <= '1';
		edo <= 20;
		
	elsif edo = 20 then
		if fin = '1' then
			minuto_out_s <= data_out;
			ini <= '0';
			edo <= 21;
		else
			ini <= '1';
		end if;	

--Se configura el protocolo I2C como lectura y se adquieren las horas.	
	elsif edo = 21 then
		reg <= x"02";
		rw <= '1';
		edo <= 22;
		
	elsif edo = 22 then
		if fin = '1' then
			hora_out_s <= data_out;
			ini <= '0';
			edo <= 23;
		else
			ini <= '1';
		end if;

--Se configura el protocolo I2C como lectura y se adquiere el da de la semana.	
	elsif edo = 23 then
		reg <= x"03";
		rw <= '1';
		edo <= 24;
		
	elsif edo = 24 then
		if fin = '1' then
			dia_semana_out_s <= data_out;
			ini <= '0';
			edo <= 25;
		else
			ini <= '1';
		end if;

--Se configura el protocolo I2C como lectura y se adquiere el dia del mes.	
	elsif edo = 25 then
		reg <= x"04";
		rw <= '1';
		edo <= 26;
		
	elsif edo = 26 then
		if fin = '1' then
			dia_out_s <= data_out;
			ini <= '0';
			edo <= 27;
		else
			ini <= '1';
		end if;

--Se configura el protocolo I2C como lectura y se adquiere el mes.	
	elsif edo = 27 then
		reg <= x"05";
		rw <= '1';
		edo <= 28;
		
	elsif edo = 28 then
		if fin = '1' then
			mes_out_s <= data_out;
			ini <= '0';
			edo <= 29;
		else
			ini <= '1';
		end if;

--Se configura el protocolo I2C como lectura y se adquiere el ao.		
	elsif edo = 29 then
		reg <= x"06";
		rw <= '1';
		edo <= 30;
		
	elsif edo = 30 then
		if fin = '1' then
			year_out_s <= data_out;
			ini <= '0';
			edo <= 31;
			LEER_FIN <= '1';
		else
			ini <= '1';
		end if;

--Se activa la bandera de FIN.	
	elsif edo = 31 then
		LEER_FIN <= '0';
		edo <= 16;
		
	end if;
end if;
end process;
	
end Behavioral;

